import importlib
import logging
import os
import signal
import sys
import time
from io import BytesIO

import gevent
import gevent.monkey
import gevent.pool
import gevent.server

from scotchwsgi import const
from scotchwsgi.response import WSGIResponseWriter
from scotchwsgi.request import WSGIRequest

logger = logging.getLogger(__name__)

class WSGIWorker(object):
    def __init__(self, app_location, sock, hostname, parent_pid, request_timeout):
        gevent.monkey.patch_all()

        # Ignore interrupts to disable KeyboardInterrupt being logged
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        # Allow app location to refer to files in cwd
        sys.path.append(os.getcwd())

        self.sock = sock
        self.hostname = hostname
        _, self.port = sock.getsockname()
        self.parent_pid = parent_pid
        self.request_timeout = request_timeout

        app_module = importlib.import_module(app_location)
        if not hasattr(app_module, 'app'):
            logger.error("'app' not found in {}".format(app_location))
            sys.exit(1)

        self.application = app_module.app

    def start(self):
        logger.info("Worker starting (PID: %d)", os.getpid())

        pool = gevent.pool.Pool(size=const.MAX_CONNECTIONS)

        server = gevent.server.StreamServer(
            self.sock,
            self._handle_connection,
            spawn=pool,
        )
        server.start()

        while True:
            if os.getppid() != self.parent_pid:
                logger.info("Worker parent changed, exiting")
                break
            time.sleep(1)

    def _handle_connection(self, conn, addr):
        logger.info("New connection: %s", addr)

        reader = conn.makefile('rb')
        writer = conn.makefile('wb')
        close_connection = False

        while not close_connection:
            try:
                with gevent.Timeout(self.request_timeout):
                    request = WSGIRequest.from_reader(reader)
            except ValueError:
                logger.error("Invalid request received from: %s", addr)
                self._send_error("400 Bad Request", writer)
                close_connection = True
            except gevent.Timeout:
                logger.info("Connection timed out: %s", addr)
                close_connection = True
            else:
                response_writer = self._send_response(request, writer)
                if not response_writer or response_writer.wrote_connection_close:
                    close_connection = True

        logger.debug("Closing connection")

        try:
            reader.close()
        except IOError:
            pass

        try:
            writer.close()
        except IOError:
            pass

        conn.close()

    def _send_response(self, request, writer):
        environ = self._get_environ(request)
        server_headers = self._get_server_headers(request)
        response_writer = WSGIResponseWriter(writer, server_headers)

        logger.debug("Calling into application")
        response_iter = self.application(environ, response_writer.start_response)
        logger.debug("Called into application")

        try:
            for response in response_iter:
                if response:
                    logger.debug("Write %s", response)
                    response_writer.write(response)

            if not response_writer.headers_sent:
                # force headers to be sent if nothing was written previously
                response_writer.write(b"")

            return response_writer
        except Exception as e:
            logger.error("Application aborted: %r", e)
        finally:
            response_iter_close = getattr(response_iter, 'close', None)
            if callable(response_iter_close):
                response_iter.close()
            logger.debug("Called into application")

    def _send_error(self, status_line, writer):
        server_headers = [('Connection', 'close')]
        response_writer = WSGIResponseWriter(writer, server_headers)
        response_writer.start_response(status_line, [])
        response_writer.write(b'')
        return response_writer

    def _get_environ(self, request):
        environ = {
            'REQUEST_METHOD': request.method,
            'SCRIPT_NAME': '',
            'SERVER_NAME': self.hostname,
            'SERVER_PORT': str(self.port),
            'SERVER_PROTOCOL': request.http_version,
            'PATH_INFO': request.path,
            'QUERY_STRING': request.query,
            'wsgi.version': (1, 0),
            'wsgi.url_scheme': 'http',
            'wsgi.input': BytesIO(request.body),
            'wsgi.errors': sys.stderr,
            'wsgi.multithread': True,
            'wsgi.multiprocess': True,
            'wsgi.run_once': False,
        }

        environ['PATH_INFO'] = request.path
        environ['QUERY_STRING'] = request.query if request.query else ''

        environ_headers = request.headers.copy()
        if 'content-type' in request.headers:
            environ['CONTENT_TYPE'] = environ_headers.pop('content-type')
        if 'content-length' in request.headers:
            environ['CONTENT_LENGTH'] = environ_headers.pop('content-length')
        for http_header_name, http_header_value in environ_headers.items():
            http_header_name = 'HTTP_{}'.format(http_header_name.upper().replace('-', '_'))
            environ[http_header_name] = http_header_value
        environ_headers.clear()

        return environ

    def _get_server_headers(self, request):
        server_headers = []
        if request.headers.get('connection', '').lower() == 'close':
            server_headers.append(('Connection', 'close'))
        return server_headers

def start_new_worker(*args, **kwargs):
    worker = WSGIWorker(*args, **kwargs)
    worker.start()
    return worker



import logging

from scotchwsgi import const

logger = logging.getLogger(__name__)

class WSGIResponseHeaders(object):
    def __init__(self, server_headers, app_headers):
        self.response_headers = []
        self.has_connection_close = False
        self.has_content_length = False

        for header_name, header_value in server_headers:
            self.response_headers.append((header_name, header_value))

            if header_name.lower() == 'connection':
                if header_value.lower() == 'close':
                    self.has_connection_close = True

        for header_name, header_value in app_headers:
            self.response_headers.append((header_name, header_value))

            if header_name.lower() == 'content-length':
                self.has_content_length = True

        if not self.has_content_length and not self.has_connection_close:
            self.response_headers.append(('Connection', 'close'))
            self.has_connection_close = True

    def __iter__(self):
        return iter(self.response_headers)

class WSGIResponseWriter(object):
    def __init__(self, writer, server_headers=None):
        self.writer = writer
        self.headers_to_send = []
        self.headers_sent = []
        self.server_headers = server_headers or []

    def start_response(self, status, app_headers, exc_info=None):
        logger.debug("start_response %s %s %s", status, app_headers, exc_info)

        if exc_info:
            try:
                if self.headers_sent:
                    logger.debug("Reraising application exception")
                    # reraise original exception if headers already sent
                    raise exc_info[1].with_traceback(exc_info[2])
            finally:
                exc_info = None # avoid dangling circular ref
        elif self.headers_to_send:
            raise AssertionError("Headers already set")

        response_headers = WSGIResponseHeaders(
            self.server_headers,
            app_headers,
        )

        self.headers_to_send[:] = [status, response_headers]

        return self.write

    def write(self, data):
        if not self.headers_to_send:
            raise AssertionError("write() before start_response()")

        elif not self.headers_sent:
            status, response_headers = self.headers_to_send[:]
            logger.debug("Send headers %s %s", status, response_headers)

            self.writer.write(b"HTTP/1.1 ")
            self.writer.write(status.encode(const.STR_ENCODING))
            self.writer.write(b"\r\n")

            for header_name, header_value in response_headers:
                self.writer.write(header_name.encode(const.STR_ENCODING))
                self.writer.write(b": ")
                self.writer.write(header_value.encode(const.STR_ENCODING))
                self.writer.write(b"\r\n")

            self.writer.write(b"\r\n")

            self.headers_sent[:] = [status, response_headers]

        self.writer.write(data)
        self.writer.flush()

    @property
    def wrote_connection_close(self):
        status_, response_headers = self.headers_sent
        return response_headers.has_connection_close



from distutils.core import setup

setup(
    name='scotchwsgi',
    version='0.2.2',
    author='Christopher Thorne',
    author_email='libcthorne@gmail.com',
    url='https://cthorne.me',
    install_requires=[
        'gevent',
    ],
    packages=['scotchwsgi'],
    scripts=[
        'bin/scotchwsgi',
    ],
)
